
/*	
	This file is part of miSCellaneous, a program library for SuperCollider 3
	
	Created: 2014-10-06, version 0.10b
	Copyright (C) 2009-2014 Daniel Mayer 
	Email: 	daniel-mayer@email.de
	URL:		www.daniel-mayer.at

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


// It might look odd that copying is triggered twice in PSx patterns, but -
// especially important with event patterns: copying within the Pfunc function ensures 
// that buffered items of the *source* are not altered and second copying 
// (delegated to the MemoRoutine) ensures that items buffered in the PSx are not 
// altered. It would of be possible to differentiate by defining two further 
// copy flags. 


PStream : Plazy { 
	var <>srcPat, <>length, <>lengthStream, <>memoRoutine;
	
	*new { |srcPat, length = inf, bufSize = 1, copyItems = 0, copySets = 1| 
		var lengthStream = length.asStream, memoRoutine;
		memoRoutine = MemoRoutine(
			{ |inval| srcPat.embedInStream(inval) }, 
			bufSize: bufSize,
			copyItems: copyItems,
			copySets: copySets
		);
		^super.new.func_({ |inval| 
			Pfinval(lengthStream.next(inval), Pfunc { memoRoutine.next(inval).miSC_copy(copyItems, copySets) }) 
		}).srcPat_(srcPat).length_(length).memoRoutine_(memoRoutine).lengthStream_(lengthStream)
	} 

	at  { |i| ^memoRoutine.lastValues[i] }

	lastValues  { ^memoRoutine.lastValues }
	
	lastValue  { ^memoRoutine.lastValues[0] }
	
	bufSize { ^memoRoutine.bufSize }

	count { ^memoRoutine.count }
} 

PS : PStream {}

Pstream : PStream {}



PSdup : PStream {
	var <>psxPat;
		
	*new { |psxPat, length = inf, bufSize = 1, copyItems = 0, copySets = 1|
		^super.new(
			psxPat.isKindOf(PStream).not.if { SimpleInitError("PSdup requires PSx source pattern").throw };
			Pfunc { psxPat.lastValue.miSC_copy(copyItems, copySets) }, length, bufSize, copyItems, copySets
		).psxPat_(psxPat);
	}
}


PSrecur : PStream {
	var <>funcStream;
	*new { |func, length = inf, bufSize, start, copyItems = 0, copySets = 1|
		var funcStr = func.asStream, lengthStream = length.asStream, memoRoutine, replaceSize;

		// adapt bufSize if not given 
		bufSize = bufSize ?? { 
			(start.isNil || (start.isKindOf(SequenceableCollection).not)).if { 1 }{ start.size };
		};
		// make start an array
		(start.notNil && (start.isKindOf(SequenceableCollection).not)).if { start = [start] };
		
		memoRoutine = MemoRoutine({ |inval| 
				Pfunc { 
					funcStr.next.(memoRoutine.lastValues, memoRoutine.count).miSC_copy(copyItems, copySets)
				}.embedInStream(inval) 
			}, 
			bufSize: bufSize,
			copyItems: copyItems,
			copySets: copySets
		);

		start.notNil.if {
			 replaceSize = min(start.size, memoRoutine.lastValues.size);
			 replaceSize.do { |i| memoRoutine.lastValues[i] = start[start.size - 1 - i] };
		};
		
		^super.new(nil, length).funcStream_(funcStr).memoRoutine_(memoRoutine)
			.func_({ |inval| Pfinval(lengthStream.next(inval), Pfunc { memoRoutine.next(inval) }) });
	}
	
}

